#pragma once
#ifndef PARSODUS_LRTABLES_LR1ITEMSETBASE_H_EREKWQSM
#define PARSODUS_LRTABLES_LR1ITEMSETBASE_H_EREKWQSM

#include "Parsodus/lrtables/LR1Item.h"
#include "Parsodus/util/symbols.h"

#include <algorithm>

namespace pds {
namespace lr {

template <typename Itemset>
class LR1ItemsetBase {
public:
    LR1ItemsetBase();
    LR1ItemsetBase(std::shared_ptr<Rule> start);

    static bool needsFollow();

    void close(const Grammar& g);
    Itemset succ(std::string sym) const;
    bool operator==(const Itemset& rhs) const;
    bool empty() const;
    std::set<std::size_t> getReduces(const Grammar& g, std::string lookahead) const;

private:
    std::set<LR1Item> m_items;
};

template <typename Itemset>
LR1ItemsetBase<Itemset>::LR1ItemsetBase()
{}

template <typename Itemset>
LR1ItemsetBase<Itemset>::LR1ItemsetBase(std::shared_ptr<Rule> start) {
    m_items.emplace(LR1Item{start, 0, {util::EOF_PLACEHOLDER}});
}

template <typename Itemset>
bool LR1ItemsetBase<Itemset>::needsFollow() {
    return true;
}

template <typename Itemset>
void LR1ItemsetBase<Itemset>::close(const Grammar& g) {
    bool changes = true;
    std::vector<LR1Item> todo;
    std::set<std::pair<std::string, std::string>> added; // (variable, lookahead)

    while (changes) {
        changes = false;
        std::set<LR1Item> toAdd;

        for (const LR1Item& i : m_items) {
            if (i.dotIdx < i.rule->tail.size()) {
                std::string& sym = i.rule->tail[i.dotIdx];
                if (g.variables.count(sym)) {
                    std::vector<std::string> seq(std::vector<std::string>(i.rule->tail.begin() + i.dotIdx + 1, i.rule->tail.end()));
                    std::set<std::string> first = (*g.first)(seq);
                    if (first.count("") || !first.size()) {
                        first.insert(i.lookaheads.begin(), i.lookaheads.end());
                        first.erase("");
                    }

                    for (std::string newLookahead : first) {
                        if (!added.count({sym, newLookahead})) {
                            added.emplace(sym, newLookahead);
                            changes = true;
                            for (const auto& rule : g.rules) {
                                if (rule->head == sym) {
                                    toAdd.insert(LR1Item{rule, 0, {newLookahead}});
                                }
                            }
                        }
                    }
                }
            }
        }

        std::vector<LR1Item> newItems;
        for (auto& it : m_items) {
            newItems.emplace_back(std::move(it));
        }
        for (const auto& newItem : toAdd) {
            bool found = false;
            for (auto& oldItem : newItems) {
                if (newItem.dotIdx == oldItem.dotIdx && newItem.rule == oldItem.rule) {
                    found = true;
                    oldItem.lookaheads.insert(newItem.lookaheads.begin(), newItem.lookaheads.end());
                    break;
                }
            }
            if (!found) {
                newItems.push_back(newItem);
            }
        }
        m_items.clear();
        for (auto& it : newItems) {
            m_items.emplace(std::move(it));
        }
    }
}

template <typename Itemset>
Itemset LR1ItemsetBase<Itemset>::succ(std::string sym) const {
    Itemset sc;
    for (auto& item : m_items) {
        if (item.dotIdx < item.rule->tail.size()) {
            if (item.rule->tail[item.dotIdx] == sym) {
                sc.m_items.insert(LR1Item{item.rule, item.dotIdx + 1, item.lookaheads});
            }
        }
    }
    return sc;
}

template <typename Itemset>
bool LR1ItemsetBase<Itemset>::operator==(const Itemset& rhs) const {
    return m_items == rhs.m_items;
}

template <typename Itemset>
bool LR1ItemsetBase<Itemset>::empty() const {
    return m_items.empty();
}

template <typename Itemset>
std::set<std::size_t> LR1ItemsetBase<Itemset>::getReduces(const Grammar& g, std::string lookahead) const {
    std::set<std::size_t> result;
    for (const auto& item : m_items) {
        if (item.dotIdx >= item.rule->tail.size() && item.lookaheads.count(lookahead)) {
            result.insert(std::find(g.rules.begin(), g.rules.end(), item.rule) - g.rules.begin());
        }
    }
    return result;
}

} /* lr */ 
} /* pds */

#endif /* PARSODUS_LRTABLES_LR1ITEMSETBASE_H_EREKWQSM */
